@@ -1,7 +1,7 @@
 ;;; exwm-manage.el --- Window Management Module for  -*- lexical-binding: t -*-
 ;;;                    EXWM
 
-;; Copyright (C) 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2016 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -27,7 +27,6 @@
 ;;; Code:
 
 (require 'exwm-core)
-(eval-when-compile (require 'exwm-workspace))
 
 (defvar exwm-manage-finish-hook nil
   "Normal hook run after a window is just managed, in the context of the
@@ -59,6 +58,20 @@ corresponding buffer.")
         (when reply
           (setq exwm--mwm-hints (append (slot-value reply 'value) nil)))))))
 
+(defvar exwm-workspace--current)
+(defvar exwm-workspace--switch-history-outdated)
+
+(declare-function exwm--update-window-type "exwm.el" (id &optional force))
+(declare-function exwm--update-class "exwm.el" (id &optional force))
+(declare-function exwm--update-transient-for "exwm.el" (id &optional force))
+(declare-function exwm--update-normal-hints "exwm.el" (id &optional force))
+(declare-function exwm--update-title "exwm.el" (id))
+(declare-function exwm--update-hints "exwm.el" (id &optional force))
+(declare-function exwm--update-protocols "exwm.el" (id &optional force))
+(declare-function exwm--update-state "exwm.el" (id &optional force))
+(declare-function exwm-floating--set-floating "exwm-floating.el" (id))
+(declare-function exwm-floating--unset-floating "exwm-floating.el" (id))
+
 (defun exwm-manage--manage-window (id)
   "Manage window ID."
   (exwm--log "Try to manage #x%x" id)
@@ -103,7 +116,7 @@ corresponding buffer.")
                             exwm-window-type))))
         (exwm--log "No need to manage #x%x" id)
         ;; Remove all events
-        (xcb:+request-checked+request-check exwm--connection
+        (xcb:+request exwm--connection
             (make-instance 'xcb:ChangeWindowAttributes
                            :window id :value-mask xcb:CW:EventMask
                            :event-mask xcb:EventMask:NoEvent))
@@ -111,13 +124,17 @@ corresponding buffer.")
         (xcb:+request exwm--connection
             (make-instance 'xcb:MapWindow :window id))
         (with-slots (x y width height) exwm--geometry
-          ;; Reparent to virtual root (essential)
-          (xcb:+request exwm--connection
-              (make-instance 'xcb:ReparentWindow
-                             :window id
-                             :parent (frame-parameter exwm-workspace--current
-                                                      'exwm-window-id)
-                             :x x :y y))
+          ;; Reparent to virtual root
+          (unless (or (memq xcb:Atom:_NET_WM_WINDOW_TYPE_DESKTOP
+                            exwm-window-type)
+                      (memq xcb:Atom:_NET_WM_WINDOW_TYPE_DOCK
+                            exwm-window-type))
+            (xcb:+request exwm--connection
+                (make-instance 'xcb:ReparentWindow
+                               :window id
+                               :parent (frame-parameter exwm-workspace--current
+                                                        'exwm-workspace)
+                               :x x :y y)))
           ;; Center window of type _NET_WM_WINDOW_TYPE_SPLASH
           (when (memq xcb:Atom:_NET_WM_WINDOW_TYPE_SPLASH exwm-window-type)
             (xcb:+request exwm--connection
@@ -126,12 +143,9 @@ corresponding buffer.")
                                :value-mask (eval-when-compile
                                              (logior xcb:ConfigWindow:X
                                                      xcb:ConfigWindow:Y))
-                               :x (/ (- (frame-pixel-width
-                                         exwm-workspace--current)
-                                        width)
+                               :x (/ (- (exwm-workspace--current-width) width)
                                      2)
-                               :y (/ (- (frame-pixel-height
-                                         exwm-workspace--current)
+                               :y (/ (- (exwm-workspace--current-height)
                                         height)
                                      2)))))
         (xcb:flush exwm--connection)
@@ -141,13 +155,35 @@ corresponding buffer.")
         (throw 'return 'ignored))
       ;; Manage the window
       (exwm--log "Manage #x%x" id)
+      ;; Create a new container as the parent of this X window
+      (setq exwm--container (xcb:generate-id exwm--connection))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:CreateWindow
+                         :depth 0 :wid exwm--container
+                         :parent (frame-parameter exwm-workspace--current
+                                                  'exwm-workspace)
+                         :x 0 :y 0 :width 1 :height 1 :border-width 0
+                         :class xcb:WindowClass:CopyFromParent
+                         :visual 0      ;CopyFromParent
+                         :value-mask (logior xcb:CW:OverrideRedirect
+                                             xcb:CW:EventMask)
+                         :override-redirect 1
+                         :event-mask xcb:EventMask:SubstructureRedirect))
+      (exwm--debug
+       (xcb:+request exwm--connection
+           (make-instance 'xcb:ewmh:set-_NET_WM_NAME
+                          :window exwm--container
+                          :data (format "EXWM container for 0x%x" id))))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:ReparentWindow
+                         :window id :parent exwm--container :x 0 :y 0))
       (xcb:+request exwm--connection    ;remove border
           (make-instance 'xcb:ConfigureWindow
                          :window id :value-mask xcb:ConfigWindow:BorderWidth
                          :border-width 0))
       (dolist (button       ;grab buttons to set focus / move / resize
                (list xcb:ButtonIndex:1 xcb:ButtonIndex:2 xcb:ButtonIndex:3))
-        (xcb:+request-checked+request-check exwm--connection
+        (xcb:+request exwm--connection
             (make-instance 'xcb:GrabButton
                            :owner-events 0 :grab-window id
                            :event-mask xcb:EventMask:ButtonPress
@@ -177,18 +213,21 @@ corresponding buffer.")
 (defun exwm-manage--unmanage-window (id &optional withdraw-only)
   "Unmanage window ID."
   (let ((buffer (exwm--id->buffer id)))
-    (exwm--log "Unmanage #x%x (buffer: %s)" id buffer)
+    (exwm--log "Unmanage #x%x (buffer: %s, widthdraw: %s)"
+               id buffer withdraw-only)
     (setq exwm--id-buffer-alist (assq-delete-all id exwm--id-buffer-alist))
-    (xcb:+request exwm--connection      ;update _NET_CLIENT_LIST
-        (make-instance 'xcb:ewmh:set-_NET_CLIENT_LIST
-                       :window exwm--root
-                       :data (vconcat (mapcar #'car exwm--id-buffer-alist))))
-    (xcb:flush exwm--connection)
     (when (buffer-live-p buffer)
       (with-current-buffer buffer
-        (when exwm--floating-frame
-          (make-frame-invisible exwm--floating-frame)
-          (redisplay))
+        ;; Flickering seems unavoidable here if the DestroyWindow request is
+        ;; not initiated by us.
+        ;; What we can do is to hide the its container ASAP.
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:UnmapWindow :window exwm--container))
+        (xcb:flush exwm--connection)
+        ;; Unmap the X window.
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:UnmapWindow :window id))
+        ;;
         (setq exwm-workspace--switch-history-outdated t)
         ;;
         (when withdraw-only
@@ -203,9 +242,7 @@ corresponding buffer.")
               (setq geometry-parent
                     (xcb:+request-unchecked+reply exwm--connection
                         (make-instance 'xcb:GetGeometry
-                                       :drawable
-                                       (frame-parameter exwm--floating-frame
-                                                        'exwm-outer-id)))
+                                       :drawable exwm--container))
                     geometry (xcb:+request-unchecked+reply exwm--connection
                                  (make-instance 'xcb:GetGeometry
                                                 :drawable id)))
@@ -221,14 +258,29 @@ corresponding buffer.")
           ;; Delete WM_STATE property
           (xcb:+request exwm--connection
               (make-instance 'xcb:DeleteProperty
-                             :window id :property xcb:Atom:WM_STATE))
-          (xcb:flush exwm--connection))
+                             :window id :property xcb:Atom:WM_STATE)))
+        (when exwm--floating-frame
+          ;; Unmap the floating frame before destroying the containers.
+          (let ((window (frame-parameter exwm--floating-frame 'exwm-outer-id)))
+            (xcb:+request exwm--connection
+                (make-instance 'xcb:UnmapWindow :window window))
+            (xcb:+request exwm--connection
+                (make-instance 'xcb:ReparentWindow
+                               :window window :parent exwm--root :x 0 :y 0))))
+        ;; Destroy the X window container (and the frame container if any).
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:DestroyWindow :window exwm--container))
         (let ((kill-buffer-query-functions nil)
               (floating exwm--floating-frame))
           (kill-buffer)
           (when floating
             (select-window
-             (frame-selected-window exwm-workspace--current))))))))
+             (frame-selected-window exwm-workspace--current)))))
+      (xcb:+request exwm--connection    ;update _NET_CLIENT_LIST
+          (make-instance 'xcb:ewmh:set-_NET_CLIENT_LIST
+                         :window exwm--root
+                         :data (vconcat (mapcar #'car exwm--id-buffer-alist))))
+      (xcb:flush exwm--connection))))
 
 (defun exwm-manage--scan ()
   "Search for existing windows and try to manage them."
@@ -239,69 +291,110 @@ corresponding buffer.")
           (xcb:+request-unchecked+reply exwm--connection
               (make-instance 'xcb:GetWindowAttributes :window i))
         (when (and (= 0 override-redirect) (= xcb:MapState:Viewable map-state))
+          (xcb:+request exwm--connection
+              (make-instance 'xcb:UnmapWindow :window i))
+          (xcb:flush exwm--connection)
           (exwm-manage--manage-window i))))))
 
 (defvar exwm-manage--ping-lock nil
   "Non-nil indicates EXWM is pinging a window.")
 (defvar exwm-manage-ping-timeout 3 "Seconds to wait before killing a client.")
 
-;;;###autoload
-(defun exwm-manage--close-window (id &optional buffer)
-  "Close window ID in a proper way."
+(defun exwm-manage--kill-buffer-query-function ()
+  "Run in `kill-buffer-query-functions'."
   (catch 'return
-    (unless (exwm--id->buffer id)
-      (throw 'return t))
-    (unless buffer (setq buffer (exwm--id->buffer id)))
-    ;; Destroy the client window if it does not support WM_DELETE_WINDOW
-    (unless (and (buffer-live-p buffer)
-                 (with-current-buffer buffer
-                   (memq xcb:Atom:WM_DELETE_WINDOW exwm--protocols)))
+    (when (xcb:+request-checked+request-check exwm--connection
+              (make-instance 'xcb:MapWindow :window exwm--id))
+      ;; The X window is no longer alive so just close the buffer.
+      ;; Destroy the container.
+      ;; Hide the container to prevent flickering.
       (xcb:+request exwm--connection
-          (make-instance 'xcb:DestroyWindow :window id))
+          (make-instance 'xcb:UnmapWindow :window exwm--container))
       (xcb:flush exwm--connection)
+      (when exwm--floating-frame
+        (let ((window (frame-parameter exwm--floating-frame 'exwm-outer-id)))
+          (xcb:+request exwm--connection
+              (make-instance 'xcb:UnmapWindow :window window))
+          (xcb:+request exwm--connection
+              (make-instance 'xcb:ReparentWindow
+                             :window window
+                             :parent exwm--root
+                             :x 0 :y 0))))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:DestroyWindow :window exwm--container))
+      (xcb:flush exwm--connection)
+      (throw 'return t))
+    (unless (memq xcb:Atom:WM_DELETE_WINDOW exwm--protocols)
+      ;; The X window does not support WM_DELETE_WINDOW; destroy it.
+      ;; Hide the container to prevent flickering.
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:UnmapWindow :window exwm--container))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:DestroyWindow :window exwm--id))
+      (xcb:flush exwm--connection)
+      ;; Wait for DestroyNotify event.
       (throw 'return nil))
-    ;; Try to close the window with WM_DELETE_WINDOW client message
+    ;; Try to close the X window with WM_DELETE_WINDOW client message.
     (xcb:+request exwm--connection
         (make-instance 'xcb:icccm:SendEvent
-                       :destination id
+                       :destination exwm--id
                        :event (xcb:marshal
                                (make-instance 'xcb:icccm:WM_DELETE_WINDOW
-                                              :window id)
+                                              :window exwm--id)
                                exwm--connection)))
     (xcb:flush exwm--connection)
-    ;; Try to determine if the client stop responding
-    (with-current-buffer buffer
-      (unless (memq xcb:Atom:_NET_WM_PING exwm--protocols)
-        ;; Ensure it's dead
-        (run-with-timer exwm-manage-ping-timeout nil
-                        `(lambda () (exwm-manage--kill-client ,id)))
-        (throw 'return nil))
-      (setq exwm-manage--ping-lock t)
+    ;;
+    (unless (memq xcb:Atom:_NET_WM_PING exwm--protocols)
+      ;; The window does not support _NET_WM_PING.  To make sure it'll die,
+      ;; kill it after the time runs out.
+      ;; Hide the container to prevent flickering.
       (xcb:+request exwm--connection
-          (make-instance 'xcb:SendEvent
-                         :propagate 0 :destination id
-                         :event-mask xcb:EventMask:NoEvent
-                         :event (xcb:marshal
-                                 (make-instance 'xcb:ewmh:_NET_WM_PING
-                                                :window id :timestamp 0
-                                                :client-window id)
-                                 exwm--connection)))
+          (make-instance 'xcb:UnmapWindow :window exwm--container))
       (xcb:flush exwm--connection)
-      (with-timeout (exwm-manage-ping-timeout
-                     (if (yes-or-no-p (format "`%s' is not responding. \
+      (run-with-timer exwm-manage-ping-timeout nil
+                      `(lambda () (exwm-manage--kill-client ,exwm--id)))
+      ;; Wait for DestroyNotify event.
+      (throw 'return nil))
+    ;; Try to determine if the X window is dead with _NET_WM_PING.
+    (setq exwm-manage--ping-lock t)
+    (xcb:+request exwm--connection
+        (make-instance 'xcb:SendEvent
+                       :propagate 0
+                       :destination exwm--id
+                       :event-mask xcb:EventMask:NoEvent
+                       :event (xcb:marshal
+                               (make-instance 'xcb:ewmh:_NET_WM_PING
+                                              :window exwm--id
+                                              :timestamp 0
+                                              :client-window exwm--id)
+                               exwm--connection)))
+    (xcb:flush exwm--connection)
+    (with-timeout (exwm-manage-ping-timeout
+                   (if (yes-or-no-p (format "'%s' is not responding. \
 Would you like to kill it? "
-                                              (buffer-name buffer)))
-                         (progn (exwm-manage--kill-client id)
-                                (throw 'return nil))
-                       (throw 'return nil)))
-        (while (and exwm-manage--ping-lock
-                    (exwm--id->buffer id)) ;may have been destroyed
-          (accept-process-output nil 0.1))))))
+                                            (buffer-name)))
+                       (progn (exwm-manage--kill-client exwm--id)
+                              ;; Kill the unresponsive X window and
+                              ;; wait for DestroyNotify event.
+                              (throw 'return nil))
+                     ;; Give up.
+                     (throw 'return nil)))
+      (while (and exwm-manage--ping-lock
+                  (exwm--id->buffer exwm--id)) ;may have been destroyed.
+        (accept-process-output nil 0.1))
+      ;; Give up.
+      (throw 'return nil))))
 
 (defun exwm-manage--kill-client (&optional id)
   "Kill an X client."
-  (interactive)
   (unless id (setq id (exwm--buffer->id (current-buffer))))
+  ;; Hide the container to prevent flickering.
+  (let ((buffer (exwm--id->buffer id)))
+    (when buffer
+      (with-current-buffer buffer
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:UnmapWindow :window exwm--container))
+        (xcb:flush exwm--connection))))
   (let* ((response (xcb:+request-unchecked+reply exwm--connection
                        (make-instance 'xcb:ewmh:get-_NET_WM_PID :window id)))
          (pid (and response (slot-value response 'value)))
@@ -321,22 +414,23 @@ Would you like to kill it? "
   (let ((obj (make-instance 'xcb:ConfigureRequest))
         buffer edges)
     (xcb:unmarshal obj data)
-    (with-slots (stack-mode window sibling x y width height border-width
-                            value-mask)
+    (with-slots (window x y width height
+                        border-width sibling stack-mode value-mask)
         obj
-      (exwm--log "ConfigureRequest from #x%x (#x%x) @%dx%d%+d%+d, border: %d"
-                 window value-mask width height x y border-width)
+      (exwm--log "ConfigureRequest from #x%x (#x%x) @%dx%d%+d%+d; \
+border-width: %d; sibling: #x%x; stack-mode: %d"
+                 window value-mask width height x y
+                 border-width sibling stack-mode)
       (if (setq buffer (exwm--id->buffer window))
           ;; Send client message for managed windows
           (with-current-buffer buffer
             (setq edges
                   (if exwm--fullscreen
                       (list 0 0
-                            (frame-pixel-width exwm-workspace--current)
-                            (frame-pixel-height exwm-workspace--current))
-                    (or exwm--floating-edges
-                        (window-inside-absolute-pixel-edges
-                         (get-buffer-window buffer t)))))
+                            (exwm-workspace--current-width)
+                            (exwm-workspace--current-height))
+                    (window-inside-absolute-pixel-edges
+                     (get-buffer-window buffer t))))
             (exwm--log "Reply with ConfigureNotify (edges): %s" edges)
             (xcb:+request exwm--connection
                 (make-instance 'xcb:SendEvent
@@ -353,14 +447,15 @@ Would you like to kill it? "
                                         :border-width 0 :override-redirect 0)
                                        exwm--connection))))
         (exwm--log "ConfigureWindow (preserve geometry)")
-        ;; Configure unmanaged windows
+        ;; Configure the unmanaged window.
         (xcb:+request exwm--connection
             (make-instance 'xcb:ConfigureWindow
                            :window window
                            :value-mask value-mask
                            :x x :y y :width width :height height
                            :border-width border-width
-                           :sibling sibling :stack-mode stack-mode)))))
+                           :sibling sibling
+                           :stack-mode stack-mode)))))
   (xcb:flush exwm--connection))
 
 (defun exwm-manage--on-MapRequest (data _synthetic)
@@ -369,7 +464,7 @@ Would you like to kill it? "
     (xcb:unmarshal obj data)
     (with-slots (parent window) obj
       (if (assoc window exwm--id-buffer-alist)
-          (exwm--log "#x%x is already managed" id)
+          (exwm--log "#x%x is already managed" window)
         (if (/= exwm--root parent)
             (progn (xcb:+request exwm--connection
                        (make-instance 'xcb:MapWindow :window window))
@@ -377,13 +472,13 @@ Would you like to kill it? "
           (exwm--log "MapRequest from #x%x" window)
           (exwm-manage--manage-window window))))))
 
-(defun exwm-manage--on-UnmapNotify (data synthetic)
+(defun exwm-manage--on-UnmapNotify (data _synthetic)
   "Handle UnmapNotify event."
-  (unless synthetic
-    (let ((obj (make-instance 'xcb:UnmapNotify)))
-      (xcb:unmarshal obj data)
-      (exwm--log "UnmapNotify from #x%x" (slot-value obj 'window))
-      (exwm-manage--unmanage-window (slot-value obj 'window) t))))
+  (let ((obj (make-instance 'xcb:UnmapNotify)))
+    (xcb:unmarshal obj data)
+    (with-slots (window) obj
+      (exwm--log "UnmapNotify from #x%x" window)
+      (exwm-manage--unmanage-window window t))))
 
 (defun exwm-manage--on-DestroyNotify (data synthetic)
   "Handle DestroyNotify event."
